/*
 *  libxcpc by Davide Libenzi (Exception and resource handling in C)
 *  Copyright (C) 2002  Davide Libenzi
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#if !defined(_XCPC_LISTS_H)
#define _XCPC_LISTS_H



#define XCM_LIST_HEAD_INIT(name) { &(name), &(name) }

#define XCM_LIST_HEAD(name) xcpc_listhead name = XCM_LIST_HEAD_INIT(name)

#define XCM_INIT_LIST_HEAD(ptr) do { \
	(ptr)->next = (ptr); (ptr)->prev = (ptr); \
} while (0)

#define XCM_INIT_LIST_LINK(ptr) do { \
	(ptr)->next = NULL; (ptr)->prev = NULL; \
} while (0)

#define XCM_LIST_ADD(new, previ, nexti) do { \
	xcpc_listhead *prev = previ; \
	xcpc_listhead *next = nexti; \
	next->prev = new; \
	(new)->next = next; \
	(new)->prev = prev; \
	prev->next = new; \
} while (0)

#define XCM_LIST_ADDH(new, head) XCM_LIST_ADD(new, head, (head)->next)
#define XCM_LIST_ADDT(new, head) XCM_LIST_ADD(new, (head)->prev, head)

#define XCM_LIST_UNLINK(previ, nexti) do { \
	(nexti)->prev = previ; \
	(previ)->next = nexti; \
} while (0)

#define XCM_LIST_DEL(entry) do { \
	XCM_LIST_UNLINK((entry)->prev, (entry)->next); \
	XCM_INIT_LIST_HEAD(entry); \
} while (0)

#define XCM_LIST_EMPTY(head) ((head)->next == head)

#define XCM_LIST_SPLICE(list, head) do { \
	xcpc_listhead *first = (list)->next; \
	if (first != list) { \
		xcpc_listhead *last = (list)->prev; \
		xcpc_listhead *at = (head)->next; \
		(first)->prev = head; \
		(head)->next = first; \
		(last)->next = at; \
		(at)->prev = last; \
	} \
} while (0)

#define XCM_LIST_COUNT(cnt, pos, head) for ((cnt) = 0, (pos) = (head)->next; \
						    (pos) != (head); \
						    (cnt)++, (pos) = (pos)->next)

#define XCM_LIST_ENTRY(ptr, typ, fld) ((typ *)((char *)(ptr) - (unsigned long)(&((typ *) 0)->fld)))
#define XCM_LIST_OFFENTRY(ptr, typ, off) ((typ *)((char *)(ptr) - (off)))
#define XCM_LIST_FOR_EACH(pos, head) for (pos = (head)->next; pos != (head); pos = (pos)->next)
#define XCM_LIST_FIRST(head) ((head)->next != (head) ? (head)->next: NULL)
#define XCM_LIST_LAST(head) ((head)->prev != (head) ? (head)->prev: NULL)
#define XCM_LIST_LINKED(ptr) ((ptr)->prev != NULL && (ptr)->next != NULL)
#define XCM_LIST_NEXT(ptr, head) ((ptr)->next != (head) ? (ptr)->next: NULL)
#define XCM_LIST_PREV(ptr, head) ((ptr)->prev != (head) ? (ptr)->prev: NULL)


#endif

